/*
 * Generate a VGA signal using a PIC32 microcontroller.
 *
 * Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "pic32_c.h"
#include "vga_display.h"



/* Display state. */

vga_display_t vga_display;



/* Initialise the state machine. */

void init_vga(display_config_t *display_config,
              void (*start_visible)(),
              void (*update_visible)(),
              void (*stop_visible)(),
              void (*vsync_high)(),
              void (*vsync_low)())
{
    /* Display parameters. */

    vga_display.display_config = display_config;

    /* Display state handlers. */

    vga_display.start_visible = start_visible;
    vga_display.update_visible = update_visible;
    vga_display.stop_visible = stop_visible;

    /* Vertical sync operations. */

    vga_display.vsync_high = vsync_high;
    vga_display.vsync_low = vsync_low;

    /* Initial state. */

    vga_display.state_handler = vbp_active;
    vga_display.line = 0;
}



/* Interrupt handlers. */

void vga_interrupt_handler(void)
{
    vga_display.line += 1;
    vga_display.state_handler();
}



/* Vertical back porch region. */

void vbp_active(void)
{
    if (vga_display.line < vga_display.display_config->visible_start)
        return;

    /* Enter the visible region. */

    vga_display.state_handler = visible_active;

    /* Set the line address. */

    vga_display.linedata = vga_display.display_config->screen_start;
    vga_display.start_visible(&vga_display);
}

/* Visible region. */

void visible_active(void)
{
    if (vga_display.line < vga_display.display_config->vfp_start)
    {
        /* Update the line address and handle wraparound. */

        if (!(vga_display.line % vga_display.display_config->line_multiplier))
        {
            vga_display.linedata += vga_display.display_config->line_length;

            if (vga_display.linedata >= vga_display.display_config->screen_limit)
                vga_display.linedata -= vga_display.display_config->screen_size;
        }

        vga_display.update_visible(&vga_display);
        return;
    }

    /* End the visible region. */

    vga_display.state_handler = vfp_active;

    /* Disable the channel for the next line. */

    vga_display.stop_visible(&vga_display);
}

/* Vertical front porch region. */

void vfp_active(void)
{
    if (vga_display.line < vga_display.display_config->vsync_start)
        return;

    /* Enter the vertical sync region. */

    vga_display.state_handler = vsync_active;

    /* Bring vsync low when the next line starts. */

    vga_display.vsync_low();
}

/* Vertical sync region. */

void vsync_active(void)
{
    if (vga_display.line < vga_display.display_config->vsync_end)
        return;

    /* Start again at the top of the display. */

    vga_display.line = 0;
    vga_display.state_handler = vbp_active;

    /* Bring vsync high when the next line starts. */

    vga_display.vsync_high();
}
